#########################################################################
##  GRID_BOOT.R
##   
##  The procedure grid_boot is designed to implement the grid bootstrap 
##  for general AR models.  
##  
##  Format: 	
##  	out <- grid_boot(dat,name,t,ar,grid,b,c,all,graph)
##  Inputs:  
##  	dat	 data vector (univariate time series)
##  	name	 name of time series, used for printed results and figures
##  	t	 trend order.  0 for no trend, 1 for constant, 2 for constant and trend
##  	ar	 autoregressive order
##  	grid	 number of gridpoints 
##  	b	 number of bootstrap replications at each gridpoint
##  	c	 level of confidence set (e.g., set c=.9 for a 90% region)
##  	all	 1 to compute confidence intervals for all AR coefficients
##  		 0 to compute confidence interval just for coefficient on Y(-1)
##  	graph	 Graph display.  0 for no graph.  1 for graph
##  
##  Note: Package "pracma" is required. Run the following command in R first: 
##        install.packages("pracma")
##
##  There are no outputs.  All results are printed to the screen
##  
##  If you have questions, please contact me at
##  
##  Bruce E. Hansen
##  Department of Economics
##  Social Science Building
##  University of Wisconsin
##  Madison, WI 53706-1393
##  behansen@wisc.edu
##  
#########################################################################

grid_boot <- function(dat,name,t,ar,grid,bq,c,all,grph){
  cc <- rbind(((1-c)/2),((c+1)/2))
  n <- nrow(dat)
  if (t==0) yt <- dat
  if (t==1) yt <- dat-mean(dat)
  if (t==2){ 
      dt <- cbind(matrix(1,n,1),as.matrix(seq(1,n,1))) 
      yt <- dat-dt%*%qr.solve(dt,dat)
  }
  y0 <- as.matrix(yt[1:ar])
  out <- est(dat,t,ar)
  y <- out$y
  x <- out$x
  betahat <- out$betahat
  sehat <- out$sehat
  e <- out$e 
  if (all==0) armax <- 1  else armax <- ar 
  g_lu <- matrix(0,armax,2)
  b_lu <- matrix(0,armax,2)
  for (k in 1:armax){
    bk <- betahat[k+t]
    sk <- sehat[k+t]
    beta0 <- as.matrix(seq((bk-5*sk),((bk-5*sk)+10*sk),10*sk/(grid-1)))
    if (t==0){ 
       kk <- as.matrix(seq((k+t+1),(ar+t),1))
    }else{
       if (k<ar){
          kk <- rbind(as.matrix(seq(1,(k+t-1),1)),as.matrix(seq((k+t+1),(ar+t),1)))
       }else{
          kk <- rbind(as.matrix(seq(1,(k+t-1),1)))
       }
    }
    xk <- x[,kk]
    tstat <- matrix(0,bq,grid)
    for (i in 1:grid){
      bi <- beta0[i]
      yk <- y-x[,(k+t)]*bi
      ai <- qr.solve(xk,yk)
      if (ar>1){
        bbi <- matrix(0,ar,1)
        bbi[k] <- bi
        if (k==1){
            bbi[seq(k+1,ar,1)] <- ai[(1+t):(ar+t-1)]
        }else{
            if (k<ar){
               bbi[rbind(as.matrix(seq(1,(k-1),1)),as.matrix(seq(k+1,ar,1)))] <- ai[(1+t):(ar+t-1)]
            }else{
               bbi[seq(1,(k-1),1)] <- ai[(1+t):(ar+t-1)]
            }
        }       
        phi <- rbind(as.matrix(bbi[2:ar]),0)-rbind(0,as.matrix(bbi[2:ar]))
        phi[1] <- phi[1]+bbi[1]
      }else{  
        phi <- bi
      }
      rr <- roots(as.vector(rbind(as.matrix((-rev(phi))),1)))
      if (min(abs(rr))<=1) y00 <- matrix(0,ar,1) else y00 <- y0
      for (j in 1:bq){
           et <- as.matrix(e[trunc(1+(n-ar)*runif(n))])
           datb <- y00
           for (ii in (nrow(y00)+1):nrow(et)){
                datbb <- et[ii]
                for (jj in 1:nrow(y00)) datbb <- datbb+phi[jj]*datb[ii-jj]
                datb <- rbind(datb,datbb)
           }
       out <- est(datb,t,ar)
       bb <- out$betahat
       sb <- out$sehat
       tstat[j,i] <- (bb[t+k]-bi)/sb[t+k]
      }
    }
    qh <- kq_reg(tstat,beta0,cc)     
    tk <- (bk-beta0)/sk
    out <- g_ends(beta0,tk,qh)
    g_l <- out$l
    g_u <- out$u
    g_lu[k,] <- cbind(g_l,g_u)
    g_lh <- height(tk,beta0,g_l)
    g_uh <- height(tk,beta0,g_u)
    b_lh <- height(qh[,2],beta0,bk)
    b_uh <- height(qh[,1],beta0,bk)
    b_l <- bk-b_lh*sk
    b_u <- bk-b_uh*sk
    b_lu[k,] <- cbind(b_l,b_u)
    if (grph==1){
    x11()
    z <- matrix(0,grid,1)
    mtit <- paste(name," AR(",ar,")",sep="")
    xtit <- paste(c*100,"% Confidence Interval for ","rho_",k,sep="")
    ytit <- "Bootstrap Quantiles"
    xtics <- range(beta0)
    ytics <- range(cbind(tk,qh))
    plot(beta0,tk,lty=1,col=1,type="l",xlim=xtics,ylim=ytics,ann=0)
    lines(beta0,z,lty=1,col=1)
    lines(beta0,qh[,2],lty=2,col=4)
    lines(beta0,qh[,1],lty=2,col=4)  
    title(main=mtit,ylab=ytit,xlab=xtit) 
    # Points #
    points(bk,0,pch=19)
    points(bk,b_uh,pch=22)
    points(bk,b_lh,pch=22)
    points(b_l,b_lh,pch=23)
    points(b_u,b_uh,pch=23)
    points(g_l,g_lh,pch=21)
    points(g_u,g_uh,pch=21)
    # Lines #
    lines(rbind(b_l,bk),rbind(b_lh,b_lh),lty=3,col=2)
    lines(rbind(bk,bk),rbind(b_lh,b_uh),lty=3,col=2)
    lines(rbind(bk,b_u),rbind(b_uh,b_uh),lty=3,col=2)
    # Arrow #
    lines(rbind(g_l,g_l),rbind(g_lh,0),lty=4,col=6)
    lines(rbind(g_u,g_u),rbind(g_uh,0),lty=4,col=6)
    lines(rbind(b_u,b_u),rbind(b_uh,0),lty=4,col=3)
    lines(rbind(b_l,b_l),rbind(b_lh,0),lty=4,col=3)
    points(g_l,0,pch=25,col=6)
    points(g_u,0,pch=24,col=6)
    points(b_u,0,pch=24,col=3)
    points(b_l,0,pch=25,col=3)   
    }
  }
  xname <- "Y(t-1)"
  for (k in 2:ar){
    kname <- paste("DY(t-",k,")",sep="")  
    xname <- rbind(xname,kname)
  }
  cat (name,"\n")
  cat ("Number of Gridpoints                ", grid,"\n")
  cat ("Number of Bootstrap Replications    ", bq,"\n")
  cat ("Confidence Level                    ", c,"\n")
  cat ("\n")
  cat ("\n") 
  cat ("Variable   ","Estimate     ","S.E.      ","\n")
  cat ("______________________________________________________________","\n")
  Txname <- format(xname,digits=4)
  Tbetahat <- format(betahat,digits=4)
  Tsehat <- format(sehat,digits=4)
  for (j in 1:nrow(xname)) cat (Txname[j],"  ",Tbetahat[t+j],"   ",Tsehat[t+j],"\n")
  cat ("______________________________________________________________","\n")
  cat ("\n")
  cat ("\n") 
  cat ("Grid-t Confidence Intervals:","\n")
  cat ("Variable  ","Lower     ","Upper     ","\n")
  cat ("______________________________________________________________","\n")
  Tg_lu <- format(g_lu,digits=4)
  for (j in 1:armax) cat (xname[j],"   ",Tg_lu[j,1],"    ",Tg_lu[j,2],"\n")
  cat ("______________________________________________________________","\n")
  cat ("\n")
  cat ("\n") 
  cat ("**********************************************************","\n")
  cat ("\n")
  cat ("\n") 
}


#########################################################################
##  PROC EST
##
##  Format: 	
##  	out <- est(dat,t,ar)
## 
##  Inputs:  
##  	dat	    nx1 data vector
##  	t	    trend order.  0 for no trend, 1 for constant, 2 for constant and trend
##  	ar	    autoregressive order
##  Outputs:
##  	y 	    nx1 vector of dependent variable
##  	x	    nx(ar+t) matrix of regressors
##  	betahat   OLS estimates
##  	sehat	    OLS standard errors
##  	e	    OLS residuals
##
#########################################################################

est <- function(dat,t,ar){
  n <- nrow(dat)
  if (t==0) x <- as.matrix(dat[ar:(n-1)])
  if (t==1) x <- cbind(matrix(1,(n-ar),1),as.matrix(dat[ar:(n-1)]))
  if (t==2) x <- cbind(matrix(1,(n-ar),1),as.matrix(seq(1,(n-ar),1)),as.matrix(dat[ar:(n-1)]))
  for (k in 2:ar) x <- cbind(x,as.matrix((dat[(ar-k+2):(n-k+1)]-dat[(ar-k+1):(n-k)])))    
  y <- as.matrix(dat[(1+ar):n])
  xxi <- solve(t(x)%*%x)  
  betahat <- xxi%*%(t(x)%*%y)
  e <- y-x%*%betahat
  sehat <- sqrt(diag(xxi)%*%(t(e)%*%e)/(n-ar-t))
list(y=y,x=x,betahat=betahat,sehat=sehat,e=e)
}

#########################################################################
##  PROC QUANTILE
##  
##  Format: 	
##  	q <- quantile(x,c)
##  Inputs:  
##  	x	nxr data matrix
##  	c	px1 list of quantiles
##  Output:
##  	q	rxp matrix of quantiles
##  
#########################################################################

quantile <- function(x,c){
  b <- nrow(x) 
  r <- ncol(x)
  xx <- x
  for (j in 1:r) xx[,j] <- sort(x[,j])
  bc <- b*c
  i <- floor(bc)
  i <- i+(i == 0)
  q <- xx[i,]*((i==bc)%*%matrix(1,1,r))+((xx[i,]+xx[(i+1),])/2)*((i!=bc)%*%matrix(1,1,r))
  out <- t(q)
  out
}

#########################################################################
##  PROC EPI
##  
##  Format: 	
##  	w <- epi(a,h,leave)
##  Inputs:  
##  	a	 nx1 vector of points to evaluate kernel
##  	h	 scalar bandwidth
##  	leave	 set to 1 to impose leave-one-out kernel estimation
##  Output:
##  	w	 nxn matrix of kernel weights
##  
#########################################################################

epi <- function(a,h,leave){
  a <- a%*%matrix(1,1,nrow(a))
  u <- (a-t(a))/h
  w <- (1-u^2)*(abs(u)<= 1)
  if (leave==1) diag(w) <- matrix(0,nrow(a),1)
  w <- w/(as.matrix(colSums(t(w)))%*%matrix(1,1,nrow(a)))
  w
}

#########################################################################
##  PROC KQ_REG
##  
##  Format: 	
##  	qh <- kq_reg(y,x,c)
##  Inputs:  
##  	y	nxr data matrix
##  	x	rx1 vector of points to evaluate kernel regression
##  	c	px1 list of quantiles
##  Output:
##  	qh	rxp matrix of kernel-smoothed quantiles
##  
#########################################################################

kq_reg <- function(y,x,c){
  hn <- 5
  cn <- nrow(c)
  hh <- as.matrix(seq(1.5,1.5+(hn-1),1)*(x[2]-x[1]))
  check <- matrix(0,hn,cn)
  q <- quantile(y,c)
  for (hi in 1:hn){
    h <- hh[hi]
    if (x[2]-x[1]<h){
       qh <- epi(x,h,1)%*%q
       check[hi,] <- colSums((q-qh)^2)
    }else{
       check[hi,] <- colSums(q^2)
    }
  }
  hi <- matrix(0,cn,1)
  for (j in 1:cn) hi[j] <- which.min(check[,j])
  h <- as.matrix(hh[hi])
  qh <- matrix(0,nrow(x),cn)
  for (ci in 1:cn) qh[,ci] <- epi(x,h[ci],0)%*%q[,ci]
  qh
}



#########################################################################
##  PROC CROSS
##  
##  Format: 	
##  	out <- cross(a,x,y)
##  Inputs:  
##  	a	rx1 list of gridpoints
##  	x	rx1 list, first line
##  	c	rx1 list, second line
##  Outputs:
##  	g1	list of upcrossings (second line intersects first from below)
##  	g2	list of downcrossings (second line intersects first from above)
##  
#########################################################################

cross <- function(a,x,y){
  ms <- 0/0
  rn <- nrow(a)
  r <- (x<y)
  d <- rbind(0,as.matrix((r[2:rn]-r[1:(rn-1)])))
  s <- as.matrix(seq(1,rn,1))
  li <- as.matrix(s[(d == 1)>0])
  if (is.nan(max(li))==0){
    al <- as.matrix(a[li-1])
    au <- as.matrix(a[li])
    bl <- as.matrix(x[li-1]-y[li-1])   
    bu <- as.matrix(y[li]-x[li])
    g1 <- (bu*al+bl*au)/(bl+bu)
  }else{
    g1 <- ms
  }
  li <- as.matrix(s[(d == -1)>0])
  if (sum(d==-1)==0) li <- 0/0
  if (is.nan(max(li))==0){
    al <- as.matrix(a[li-1])
    au <- as.matrix(a[li])
    bl <- as.matrix(x[li-1]-y[li-1])   
    bu <- as.matrix(y[li]-x[li])
    g2 <- (bu*al+bl*au)/(bl+bu)
  }else{
    g2 <- ms
  }
  list(g1=g1,g2=g2)
}

#########################################################################
##  PROC G_ENDS
##  
##  Format: 	
##  	out <- g_ends(a0,ta,grid)
##  Inputs:  
##  	a0	rx1 list of gridpoints
##  	ta	rx1 test statistic line
##  	grid	rx2 lower and upper quantiles
##  Output:
##  	l	lower endpoint of confidence interval
##  	u	upper endpoint of confidence interval
##  
#########################################################################

g_ends <- function(a0,ta,grid){
  out <- cross(a0,ta,as.matrix(grid[,1]))
  l1 <- out$g1
  l2 <- out$g2
  out <- cross(a0,ta,as.matrix(grid[,2]))
  u1 <- out$g1
  u2 <- out$g2
  s <- as.matrix(sort(rbind(l1,l2,u1,u2)))
  l <- min(s)
  u <- max(s)
  if (sum(is.nan(u2)==0)>0) if (l==min(u2)) l <- min(a0)
  if (sum(is.nan(l1)==0)>0) if (l==min(l1)) l <- min(a0)
  if (sum(is.nan(l2)==0)>0) if (u==max(l2)) u <- max(a0)
  if (sum(is.nan(u1)==0)>0) if (u==max(u1)) u <- max(a0)
  list(l=l,u=u)
}

#########################################################################
##  PROC HEIGHT
##  
##  Format: 	
##  	qh <- height(q,a,l)
##  Inputs:  
##  	q	rx1 heights at gridpoints
##  	a	rx1 gridpoints
##  	l	scalar, point to evaluate q
##  Output:
##  	qh	scalar, value of q at point l
##  
#########################################################################

height <- function(q,a,l){
  l2 <- which.max(a>l)
  l1 <- l2-1
  q1 <- q[l1]
  q2 <- q[l2]
  w1 <- a[l2]-l
  w2 <- l-a[l1]
  qh <- (w1*q1+w2*q2)/(w1+w2)
  qh
}


